from .djangomcp import global_mcp_server as mcp_server
from .djangomcp import MCPToolset
from .query_tool import ModelQueryToolset
from .djangomcp import (drf_serialize_output, drf_publish_create_mcp_tool,
                        drf_publish_update_mcp_tool, drf_publish_destroy_mcp_tool,
                        drf_publish_list_mcp_tool)


from django.core.management.base import BaseCommand
from mcp_server import mcp_server


class Command(BaseCommand):
    help = 'Run the global mcp server over STDIO transport'

    def handle(self, *args, **options):
        mcp_server.run(transport="stdio")


from django.conf import settings
from django.http import HttpResponse
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from mcp.server import FastMCP
from rest_framework.views import APIView

from mcp_server.djangomcp import global_mcp_server


@method_decorator(csrf_exempt, name='dispatch')
class MCPServerStreamableHttpView(APIView):
    mcp_server = global_mcp_server

    def get(self, request, *args, **kwargs):
        return self.mcp_server.handle_django_request(request)

    def post(self, request, *args, **kwargs):
        return self.mcp_server.handle_django_request(request)

    def delete(self, request, *args, **kwargs):
        self.mcp_server.destroy_session(request)
        return HttpResponse(status=200, content="Session destroyed")



"""
URL configuration for mcpexample project.

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/5.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.contrib import admin
from django.urls import path, include
from django.utils.module_loading import import_string
from rest_framework.permissions import IsAuthenticated

from mcp_server.views import MCPServerStreamableHttpView


# Register MCP Server View and bypass default DRF default permission / authentication classes
urlpatterns = [
    path("mcp", MCPServerStreamableHttpView.as_view(
        permission_classes=[IsAuthenticated] if getattr(settings, 'DJANGO_MCP_AUTHENTICATION_CLASSES', None) else [],
        authentication_classes=[import_string(cls) for cls in getattr(settings, 'DJANGO_MCP_AUTHENTICATION_CLASSES', [])]
    ), name="mcp_server_streamable_http_endpoint"),
]



import base64
import csv
import io
import json
import logging
from random import randint

from asgiref.sync import sync_to_async
from django.conf import settings
from django.db import models
from django.db.models import Q, QuerySet, Count, Sum, Model
from django.db.models import CharField, TextField
from django.db.models import Avg, Max, Min
from django.utils.module_loading import import_string
from mcp.types import EmbeddedResource, TextResourceContents, BlobResourceContents, TextContent
from rest_framework.renderers import BaseRenderer

logger = logging.getLogger(__name__)

"""
    Tools to generate MongoDB-style $jsonSchema from Django models
    and to apply JSON-like queries to Django QuerySets using a subset 
    of MangoDB aggregation pipeline syntax.
"""


def generate_json_schema(model, fields=None, exclude=None):
    """
    Generate a MongoDB-style $jsonSchema from a Django model.

    Args:
        model: Django model class
        fields: Optional list of field names to include
        exclude: Optional list of field names to exclude

    Returns:
        A dict representing the $jsonSchema
    """

    type_mapping = {
        models.CharField: "string",
        models.TextField: "string",
        models.IntegerField: "int",
        models.FloatField: "double",
        models.BooleanField: "bool",
        models.DateTimeField: "date",
        models.DateField: "date",
        models.TimeField: "string",
        models.EmailField: "string",
        models.URLField: "string",
        models.DecimalField: "double",
        models.AutoField: "int",
        models.BigAutoField: "long",
        models.BigIntegerField: "long",
        models.JSONField: "object",
    }

    schema = {
        "description": (model.__doc__ or "").strip(),
        "$jsonSchema": {
            "bsonType": "object",
            "properties": {},
            "required": []
        }
    }

    for field in model._meta.get_fields():
        if not field.concrete:
            continue
        if fields and field.name not in fields:
            continue
        if exclude and field.name in exclude:
            continue

        prop = {}

        # Primary key description
        if getattr(field, 'primary_key', False):
            prop["description"] = "Primary unique identifier for this model"

        # ForeignKey
        if isinstance(field, models.ForeignKey):
            prop["bsonType"] = "objectId"
            prop["description"] = f"Reference to {field.related_model.__name__}"
            if field.help_text:
                prop["description"] += ": " + str(field.help_text)
            prop["ref"] = field.related_model.__name__
        else:
            # Type detection
            for django_type, bson_type in type_mapping.items():
                if isinstance(field, django_type):
                    prop["bsonType"] = bson_type
                    break
            else:
                prop["bsonType"] = "string"

            # Regular field description
            if field.help_text:
                prop["description"] = field.help_text
            if field.choices:
                # Add enum values
                prop["enum"] = [choice[0] for choice in field.choices]

                # Build display labels
                choice_desc = ", ".join(f"{repr(val)} = {label}" for val, label in field.choices)

                # Append to existing or new description
                if "description" in prop:
                    prop["description"] += f" Choices: {choice_desc}"
                else:
                    prop["description"] = f"Choices: {choice_desc}"

        schema["$jsonSchema"]["properties"][field.name] = prop

        if not getattr(field, 'null', True) and not getattr(field, 'blank', True):
            schema["$jsonSchema"]["required"].append(field.name)

    if not schema["$jsonSchema"]["required"]:
        del schema["$jsonSchema"]["required"]

    return schema



PIPELINE_DSL_SPEC="""
The syntax to query is a subset of MangoDB aggregation pipeline JSON with support of following stages : 

1. $lookup: Joins another collection :.
  - "from" must refer to a model name listed in ref in the schema (if defined).
  - "localField" must be a field path on the base colletion or a previous $lookup alias.
  - "foreignField" must be "_id"
  - "as" defines an alias used in subsequent $match and $lookup stages as a prefix (e.g., alias.field).
2. $match: Filter documents using comparison and logical operators.
  - Supports: $eq, $ne, $gt, $gte, $lt, $lte, $in, $nin, $regex in addition to $text for collections that support full text search.
  - Field references can include lookup aliases via dot notation, e.g. "user.name"
3. $sort: Sorts the result. Keys must map to model fields.
4. $limit: Truncates the result set to the specified number of items.
5. $project: Selects specific fields for results. Only "flat" objects are supported.
   Value is either a number/boolean to include/exclude the field or a string starting in format 
   "$<lookupAlias>.<field>" to project a field from a previous $lookup stage.  
6. $search: For collection that support full-text search. Limited to {"text":{"query":"<keyword>"}}.
7. $group: Groups the result set by a field and applies aggregations.
     - It must be the **final** stage in the pipeline.
     - You cannot have a $project stage in the pipeline.
     - `_id` can be null for global aggregation or a $<field> reference of a single field or lookup field or an object mapping "keys" to "$<field>" refs.
     - Supported accumulator operators: `$sum`, `$avg`, `$min`, `$max` and `$count`

All other stages NOT SUPPORTED : $addFields, $set, $unset, $unwind ...
"""

def apply_json_mango_query(queryset: QuerySet, pipeline: list[dict],
                           allowed_models: list = None, extended_operators: list = None,
                           text_search_fields: list | str = '*'):
    """
    Apply a JSON-like query to a Django QuerySet using a subset of MangoDB aggregation pipeline syntax.
    see pipeline_dsl_spec() for details.
    :param queryset: The base queryset to query
    :param pipeline: a list of stages to apply to the queryset compliant with pipeline_dsl_spec()
    :param allowed_models: List of allowed models for $lookup stages. If None, all models are allowed. Can be the string name or the Model class.
    :param extended_operators: List of Queryset API lookups to support as exetended operators. this interprets {"<field>":{"$<op>": value} as Q({field}__{op}=value)
    :param text_search_fields: List of field names to apply `$text` full-text search to. Use "*" to apply to all CharField and TextField fields of the model. Required if `$text` is used.
    :return: an iterable (eventually the queryset) of JSON results.
    """

    if extended_operators is None:
        extended_operators = []

    if allowed_models:
        allowed_models = [model.lower() if isinstance(model, str) else model._meta.model_name.lower()  for model in allowed_models]

    model = queryset.model
    if text_search_fields == "*":
        text_search_fields = [f.name for f in model._meta.get_fields() if
                              isinstance(f, (CharField, TextField)) and f.concrete and not f.is_relation]

    lookup_alias_map = {}

    # First pass: Apply lookups
    for stage in pipeline:
        if "$lookup" in stage:
            lookup = stage["$lookup"]
            _validate_lookup(model, lookup, allowed_models, lookup_alias_map)
            as_field = lookup["as"]
            local_field = _translate_field(lookup["localField"], lookup_alias_map)
            foreign_field = lookup["foreignField"]
            lookup_alias_map[as_field] = {
                "prefix": local_field.replace("_id", ""),
                "foreign_field": foreign_field
            }

    # Second pass: Apply rest
    projection_fields = None
    projection_mapping = None
    skip_value = None

    for i, stage in enumerate(pipeline):
        if "$match" in stage:
            match_stage = stage["$match"]
            if "$text" in match_stage:
                if not text_search_fields:
                    raise ValueError("$text used but full text search is not supported for this collection.")
                search_value = match_stage["$text"].get("$search", "")
                del match_stage["$text"]
                q = _build_text_search_q(search_value, text_search_fields)
                if match_stage:
                    q &= _parse_match(match_stage, extended_operators, lookup_alias_map, text_search_fields)
                queryset = queryset.filter(q)
            else:
                queryset = queryset.filter(_parse_match(stage["$match"], extended_operators, lookup_alias_map, text_search_fields=[]))

        elif "$search" in stage:
            search = stage["$search"]
            if not text_search_fields:
                raise ValueError("$search used but full text search is not supported for thsi collection.")
            search_value = search["text"]["query"]
            path = search["text"].get("path", text_search_fields)
            search_fields = [path] if isinstance(path, str) else path

            if not all(f in text_search_fields for f in search_fields):
                raise ValueError("$search path contains fields that are not allowed for search")
            q = _build_text_search_q(search_value, search_fields)
            queryset = queryset.filter(q)

        elif "$sort" in stage:
            order = []
            for field, direction in stage["$sort"].items():
                order.append(field if direction == 1 else f"-{field}")
            queryset = queryset.order_by(*order)

        elif "$skip" in stage:
            skip_value = stage["$skip"]

        elif "$limit" in stage:
            queryset = queryset[:stage["$limit"]]

        elif "$project" in stage:
            if any("$group" in s for s in pipeline[i+1:]):
                raise ValueError("$project cannot appear when pipeline contains $group :"
                                 " please review pipeline syntax constriants.")
            projection_fields, projection_mapping = _interpret_projection(stage["$project"], lookup_alias_map)

        elif "$group" in stage:
            if i != len(pipeline) - 1:
                raise ValueError("$group must be the last stage in the pipeline.:"
                                 " please review pipeline syntax constriants.")

            group = stage["$group"]
            group_id = group["_id"]
            annotations = {}

            for key, agg in group.items():
                if key == "_id":
                    continue
                if not isinstance(agg, dict) or len(agg) != 1:
                    raise ValueError(f'Aggregation for key {key} can only be a JSON object of format '+'{"$<operator>": "<parameter>"}.')
                op, arg = next(iter(agg.items()))
                if op == "$sum":
                    if arg == 1:
                        annotations[key] = Count("id")
                    elif isinstance(arg, str) and arg.startswith("$"):
                        annotations[key] = Sum(_translate_field(arg[1:], lookup_alias_map))
                    else:
                        raise ValueError("$sum only supports 1 or field references.")
                elif op == "$avg":
                    if isinstance(arg, str) and arg.startswith("$"):
                        annotations[key] = Avg(_translate_field(arg[1:], lookup_alias_map))
                    else:
                        raise ValueError("$avg requires a field reference.")
                elif op == "$min":
                    if isinstance(arg, str) and arg.startswith("$"):
                        annotations[key] = Min(_translate_field(arg[1:], lookup_alias_map))
                    else:
                        raise ValueError("$min requires a field reference.")
                elif op == "$max":
                    if isinstance(arg, str) and arg.startswith("$"):
                        annotations[key] = Max(_translate_field(arg[1:], lookup_alias_map))
                    else:
                        raise ValueError("$max requires a field reference.")

                elif op == "$count":
                    if arg == 1:
                        annotations[key] = Count("id")
                    elif isinstance(arg, str) and arg.startswith("$"):
                        annotations[key] = Count(_translate_field(arg[1:], lookup_alias_map))
                    else:
                        raise ValueError("$count only supports value 1 or a field reference.")
                else:
                    raise ValueError(f"Unsupported aggregation operator: {op}")

            if group_id is None:
                return [queryset.aggregate(**annotations)]
            elif isinstance(group_id, str) and group_id.startswith("$"):
                group_field = _translate_field(group_id[1:], lookup_alias_map)
                mapping = dict((k,k) for k in group.keys() if k != "_id")
                mapping[group_id[1:]] = group_field

                return _postprocess_projection(queryset.values(group_field).annotate(**annotations), mapping)
            elif isinstance(group_id, dict):
                mapping = dict((k,k) for k in group.keys() if k != "_id")
                group_fields = set()
                for key, formula in group_id.items():
                    if not isinstance(formula, str) or not formula.startswith("$"):
                        raise ValueError(f"_id in $group only supports null, string or one level object: value of {key} must be a string reference to field like $<field>.")
                    group_field = _translate_field(formula[1:], lookup_alias_map)
                    mapping["_id."+key] = group_field
                    group_fields.add(group_field)
                return _postprocess_projection(queryset.values(*group_fields).annotate(**annotations), mapping)
            else:
                raise ValueError("Unsupported _id value in $group. Only allowed : null a \"$field\" or a {\"key\":\"$field\",...} object.")

        elif "$lookup" in stage:
            continue

        else:
            raise ValueError(f"Unsupported stage {stage} : please review pipeline syntax constraints")

    if skip_value is not None:
        queryset = queryset[skip_value:]

    if projection_fields:
        queryset = queryset.values(*projection_fields)
        return _postprocess_projection(queryset, projection_mapping)

    return _postprocess_projection(queryset.values(), None)


def _interpret_projection(projection, lookup_map):
    fields = []
    mapping = {}
    for output_field, spec in projection.items():
        if isinstance(spec, str) and spec.startswith("$"):
            path = spec[1:]
            if path == "_id":
                path = "pk"
            internal_field = _translate_field(path, lookup_map)
            fields.append(internal_field)
            mapping[output_field] = internal_field
        elif spec:
            path = output_field if output_field != "_id" else "pk"
            internal_field = _translate_field(path, lookup_map)
            fields.append(internal_field)
            mapping[output_field] = internal_field
    return fields, mapping


def _postprocess_projection(queryset, projection_mapping):
    if not projection_mapping:
        yield from queryset
        return

    for row in queryset:
        result = {}
        for key, internal_key in projection_mapping.items():
            value = row.get(internal_key)
            _assign_nested_value(result, key.split("."), value)
        yield result


def _assign_nested_value(target, path_parts, value):
    for part in path_parts[:-1]:
        target = target.setdefault(part, {})
    target[path_parts[-1]] = value


def _restore_field_path(field, lookup_map):
    for alias, info in lookup_map.items():
        prefix = info['prefix']
        if field.startswith(prefix + "__"):
            return alias + "." + field[len(prefix + "__"):].replace("__", ".")
    return field.replace("__", ".")


def _validate_lookup(model, lookup, allowed_models, lookup_map):
    from_model_name = lookup["from"]
    if allowed_models is not None and from_model_name.lower() not in allowed_models:
        raise ValueError(f"Invalid lookup from collection '{from_model_name}'  : please reveiw schemas.")

    local_field_name = _translate_field(lookup["localField"], lookup_map)
    foreign_field_name = lookup["foreignField"]

    base_model, field_name = _resolve_model_from_path(model, local_field_name, lookup_map)

    try:
        local_field = base_model._meta.get_field(field_name)
    except Exception:
        raise ValueError(f"Invalid localField '{lookup['localField']}' : please review supported pipeline syntax and  schemas for {base_model._meta.model_name}.'.")

    if not local_field.is_relation or not local_field.many_to_one:
        raise ValueError(f"Invalid localField '{lookup['localField']}' : is not a reference, please review supported pipeline syntax and  schemas for {base_model._meta.model_name}.")

    related_model = local_field.related_model
    if foreign_field_name != related_model._meta.pk.name and foreign_field_name != "_id" and foreign_field_name != "pk":
        raise ValueError(f"Invalid foreignField '{lookup['foreignField']}': please review supported pipeline syntax and  schemas for {base_model._meta.model_name}.")


def _resolve_model_from_path(model, field_path, lookup_map):
    parts = field_path.split("__")
    current_model = model
    for part in parts[:-1]:
        try:
            field = current_model._meta.get_field(part)
            if field.is_relation:
                current_model = field.related_model
            else:
                break
        except Exception:
            raise ValueError(f"Invalid field path '{field_path}' at '{part}' in model '{current_model.__name__}'.")
    return current_model, parts[-1]


def _parse_match(match, extended_operators, lookup_map, text_search_fields=None):
    if "$and" in match:
        return Q(*[_parse_match(cond, extended_operators, lookup_map) for cond in match["$and"]])
    if "$or" in match:
        return Q(*[_parse_match(cond, extended_operators, lookup_map) for cond in match["$or"]], _connector=Q.OR)
    if "$nor" in match:
        return ~Q(*[_parse_match(cond, extended_operators, lookup_map) for cond in match["$nor"]], _connector=Q.OR)

    q = Q()
    for field, condition in match.items():
        field = _translate_field(field, lookup_map)

        if isinstance(condition, dict):
            for op, value in condition.items():
                if op.startswith("$"):
                    op_name = op[1:]
                    negate=False
                    if op_name == "ne":
                        negate = True
                        op_name = "eq"
                    elif op_name == "nin":
                        negate = True
                        op_name = "in"

                    if op_name == "eq" and value is None:
                        key = f"{field}__isnull"
                        value = True
                    elif op_name in ["eq", "gt", "gte", "lt", "lte", "in"]:
                        django_op = "" if op_name=="eq" else f"__{op_name}"
                        key = f"{field}{django_op}"
                    elif op_name == "regex":
                        key = f"{field}__regex"
                    elif op_name in extended_operators:
                        key = f"{field}__{op_name}"
                    else:
                        raise ValueError(f"Unsupported operator {op} : please reveiwe pipeline syntax constraints ")
                    q &= ~Q(**{key: value}) if negate else Q(**{key: value})
                else:
                    raise ValueError(f"Unknown match key: {op} : please reveiwe pipeline syntax constraints")
        else:
            q &= Q(**{field: condition})
    return q


def _translate_field(field, lookup_map):
    if field == "_id":
        return "pk"
    if "." in field:
        alias, rest = field.split(".", 1)
        if alias in lookup_map:
            return f"{lookup_map[alias]['prefix']}__{rest}"
        else:
            raise ValueError(f"Unknown lookup alias '{alias}', ensure it appears in the 'as' field of a previous $lookup")
    return field


def _build_text_search_q(search_value, fields):
    q = Q()
    for word in search_value.strip().split():
        word_q = Q()
        for f in fields:
            word_q |= Q(**{f"{f}__icontains": word})
        q &= word_q
    return q


class ModelQueryToolsetMeta(type):
    registry = {}

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        if name != "ModelQueryToolset":
            ModelQueryToolsetMeta.registry[name] = cls


class ModelQueryToolset(metaclass=ModelQueryToolsetMeta):
    """
    Base class for models that can be queried using the MCP QueryTool.
    """

    mcp_server: 'DjangoMCP' = None
    "The server to use, if not set, the global one will be used."

    model: type(Model) = None
    " The model to query, this is used to generate the JSON schema and the query methods. "

    exclude_fields: list[str] = []
    """List of fields to exclude from the schema. Related fields to collections that are not published"
    in any other ModelQueryToolset of same server will be autoamtically excluded"""

    fields: list[str] = None
    "The list of fields to include"

    search_fields: list[str] = None
    "List of fields for full text search, if not set it defaults to textual fields allowed by 'fields' parameters."

    extra_filters: list[str] = None
    "A list of queryset api filters that will be accessible to the MCP client for querying."

    extra_instructions: str = None
    "Extra instruction to provide, for example on when to query this collection"

    output_format : str = "json"
    "Desired output format, the corresponding DRF renderer class must be registered in the DJANGO_MCP_OUTPUT_RENDERER_CLASSES setting. By default JSONRenderer is used."

    output_as_resource = False
    "When set to true the MCP result will return the result as an embedded resource"

    @classmethod
    def get_text_search_fields(cls):
        if hasattr(cls, "_effective_text_search_fields"):
            return cls._effective_text_search_fields
        if cls.search_fields is not None:
            cls._effective_text_search_fields = set(cls.search_fields)
        elif cls.fields is None:
            cls._effective_text_search_fields = set(f.name for f in cls.model._meta.get_fields() if
                                                    isinstance(f, (
                                                    CharField, TextField)) and f.concrete and not f.is_relation)
        else:
            model_fields = cls.model._meta.get_fields()
            cls._effective_text_search_fields = set(f for f in cls.fields if not model_fields[f].is_relation and
                                                    isinstance(model_fields[f], (CharField, TextField)))
        if not cls._effective_text_search_fields:
            logger.debug(f"Full text search disabled for {cls.model}: no search fields resolved")
        else:
            logger.debug(
                f"Full text search for {cls.model} enabled on fields: {','.join(cls._effective_text_search_fields)}")
        return cls._effective_text_search_fields

    @classmethod
    def get_published_models(cls):
        if hasattr(cls, "_effective_published_models"):
            return cls._effective_published_models
        cls._effective_published_models = set(c.model for c in ModelQueryToolsetMeta.registry.values() if
                                              c.mcp_server == cls.mcp_server)
        return cls._effective_published_models

    @classmethod
    def get_excluded_fields(cls):
        if hasattr(cls, "_effective_excluded_fields"):
            return cls._effective_excluded_fields
        cls._effective_excluded_fields = set(cls.exclude_fields or [])
        published_models = cls.get_published_models()
        unpublished_fks = [f.name for f in cls.model._meta.get_fields()
                           if f.is_relation and f.related_model not in published_models]
        if unpublished_fks:
            logger.info(f"The following related fields of {cls.model} will not be published in {cls} "
                        f"because their models are not published: {unpublished_fks}")
            cls._effective_excluded_fields.update(
                unpublished_fks
            )
        return cls._effective_excluded_fields

    def get_queryset(self) -> QuerySet:
        """
        Returns the queryset to use for this toolset. This method can be overridden to filter the queryset
        based on the request or other parameters, the request is available in self.request and the MCP Server
        context in self.context
        """
        return self.model._default_manager.all()

    def __init__(self, context=None, request=None):
        self.context = context
        self.request = request


class _QueryExecutor:
    def __init__(self, query_tool_models, context=None, request=None):
        self.query_tool_models = query_tool_models
        self.context = context
        self.request = request

    def query(self, collection : str, search_pipeline: list[dict] = []):
        mql_model = self.query_tool_models.get(collection.lower())
        if not mql_model:
            raise ValueError(f"No such collection, available collections: {', '.join(self.query_tool_models.keys())}")
        instance = mql_model(self.context, self.request)
        qs = instance.get_queryset()

        ret = list(apply_json_mango_query(qs, search_pipeline,
                                           text_search_fields=instance.get_text_search_fields(),
                                           allowed_models=instance.get_published_models(),
                                           extended_operators=instance.extra_filters))

        if not ret:
            if instance.output_as_resource:
                return ["No results found"]
            else:
                return ret

        renderer = _output_formats.get(instance.output_format)

        assert renderer is not None, "Bad output format, but we were meant to have validated it at startup"

        if not isinstance(renderer, BaseRenderer):
            renderer = renderer()
        ret = renderer.render(ret)

        if instance.output_as_resource:
            if not ret: return ["No results found"]
            if (renderer.media_type.startswith("application/") and  "json" in renderer.media_type) \
                    or renderer.media_type.startswith("text/"):
                return ["Results attached", EmbeddedResource(
                    type="resource",
                    resource=TextResourceContents(
                        uri=f"resource://query_result/{renderer.format}",
                        mimeType=renderer.media_type,
                        text=ret
                    )
                )]
            else:
                return ["Results attached", EmbeddedResource(
                    type="resource",
                    resource=BlobResourceContents(
                        uri=f"resource://query_result/{renderer.format}",
                        mimeType=renderer.media_type,
                        blob=base64.b64encode(ret).decode('utf-8')
                    )
                )]

        else:
            return [TextContent(type="text", text=ret)]



class QueryTool:
    def __init__(self):
        self.query_tool_models = {}

    def add_query_tool_model(self, cls):
        if cls.output_format not in _output_formats:
            raise ValueError(f"Output format {cls.output_format} is not supported, available formats: {', '.join(_output_formats.keys())}. Verify your DJANGO_MCP_OUTPUT_RENDERER_CLASSES setting.")
        self.query_tool_models[cls.model._meta.model_name.lower()] = cls

    def get_instructions(self):
        ret = f"""
Use this tool to query data available in the server. 
The `collection` parameter specifies the collection to query and the `search_pipeline` parameter is 
a list of stage of a MongoDB aggregation pipeline with restricted syntax.

## MongoDB aggregation pipeline syntax supported
{PIPELINE_DSL_SPEC}. 

## Available collections to query
"""
        for name, cls in self.query_tool_models.items():
            ret+=f"""
### '{name}' collection
Documents conform the following JSON Schema
```json
{generate_json_schema(cls.model, fields=cls.fields,
                      exclude=cls.get_excluded_fields())}
```

"""
            if cls.get_text_search_fields():
                ret += "Full text search is supported on the following fields: " + ", ".join(
                    cls.get_text_search_fields()) + "."
            else:
                ret += "Full text search is not supported on this collection."

            if cls.extra_instructions:
                ret += f"\n\nExtra instructions for this collection:\n\n{cls.extra_instructions}\n"
        return ret

    def executor_factory(self, context, request):
        return _QueryExecutor(self.query_tool_models, context, request)

    def _add_tools_to(self, tool_manager):
        from .djangomcp import _ToolsetMethodCaller
        # ITerate all the methods whose name does not start with _ and register them with mcp_server.add_tool
        ret = []
        def _dumb_query(collection : str, search_pipeline: list[dict] = []):
            ...

        name = "query_data_collections"

        tool = tool_manager.add_tool(
            fn=sync_to_async(_dumb_query),
            name=name,
            description=self.get_instructions()
        )

        tool.context_kwarg = "_context"
        tool.fn = _ToolsetMethodCaller(self.executor_factory, "query", "_context", False)
        return [tool]

_output_formats=None

def init(global_mcp_server : 'DjangoMCP'):
    global _output_formats
    renderers_classes = (import_string(renderer_class) for renderer_class in
                           getattr(settings, "DJANGO_MCP_OUTPUT_RENDERER_CLASSES", ["rest_framework.renderers.JSONRenderer"]))
    _output_formats = {renderer_class.format: renderer_class for renderer_class in renderers_classes}

    server_tools = {}
    for name, cls in ModelQueryToolsetMeta.registry.items():
        cls.mcp_server = cls.mcp_server or global_mcp_server
        querytool = server_tools.get(cls.mcp_server)
        if not querytool:
            querytool = QueryTool()
            server_tools[cls.mcp_server] = querytool

        querytool.add_query_tool_model(cls)

    for server, tool in server_tools.items():
        server.register_mcptoolset(tool)





import contextvars
import functools
import inspect
import json
import logging
from importlib import import_module
from io import BytesIO
from types import SimpleNamespace
from typing import TYPE_CHECKING, Type, Callable

from asgiref.sync import sync_to_async, async_to_sync
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import QuerySet
from django.http import HttpResponse, HttpRequest
from mcp.server import FastMCP
from mcp.server.streamable_http_manager import StreamableHTTPSessionManager
from rest_framework.mixins import CreateModelMixin, UpdateModelMixin, DestroyModelMixin, ListModelMixin
from rest_framework.serializers import Serializer
from starlette.datastructures import Headers
from starlette.types import Scope, Receive, Send

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__name__)

django_request_ctx = contextvars.ContextVar("django_request")



def drf_serialize_output(serializer_class: type[Serializer]):
    """
    This annotation will process the tool result thorugh the given DRF serializer

    ```
    @drf_serialize_output(MyDRFSerializer)
    def my_function(args):
        return MyInstance()
    ```


    :param serializer_class:
    :return:
    """

    def annotator(fn):
        fn.__dmcp_drf_serializer = serializer_class
        return fn

    return annotator


class _SyncToolCallWrapper:
    def __init__(self, fn):
        self.fn = fn
        functools.update_wrapper(self, fn)

    def __call__(self, *args, **kwargs):
        try:
            ret = self.fn(*args, **kwargs)
        except:
            # TODO create kind of exception like "ToolError" that is logged only debug
            logger.exception("Error in tool invocation")
            raise
        if isinstance(ret, QuerySet):
            ret = list(ret)
        serializer_class = getattr(self.fn, '__dmcp_drf_serializer', None)
        if serializer_class is not None:
            ret = serializer_class(ret).data
        return ret


async def _call_starlette_handler(django_request: HttpRequest, session_manager: StreamableHTTPSessionManager):
    """
    Adapts a Django request into a Starlette request and calls session_manager.handle_request.

    Returns:
        A Django HttpResponse
    """
    django_request_ctx.set(django_request)
    body = json.dumps(django_request.data, cls=DjangoJSONEncoder).encode("utf-8")

    # Build ASGI scope
    scope: Scope = {
        "type": "http",
        "http_version": "1.1",
        "method": django_request.method,
        "headers": [
                       (key.lower().encode("latin-1"), value.encode("latin-1"))
                       for key, value in django_request.headers.items() if key.lower() != "content-length"
                   ] + [("Content-Length", str(len(body)).encode("latin-1"))],
        "path": django_request.path,
        "raw_path": django_request.get_full_path().encode("utf-8"),
        "query_string": django_request.META["QUERY_STRING"].encode("latin-1"),
        "scheme": "https" if django_request.is_secure() else "http",
        "client": (django_request.META.get("REMOTE_ADDR"), 0),
        "server": (django_request.get_host(), django_request.get_port()),
    }

    async def receive() -> Receive:
        return {
            "type": "http.request",
            "body": body,
            "more_body": False,
        }

    # Prepare to collect send events
    response_started = {}
    response_body = bytearray()

    async def send(message: Send):
        if message["type"] == "http.response.start":
            response_started["status"] = message["status"]
            response_started["headers"] = Headers(raw=message["headers"])
        elif message["type"] == "http.response.body":
            response_body.extend(message.get("body", b""))

    async with session_manager.run():
        # Call transport
        await session_manager.handle_request(scope, receive, send)

    # Build Django HttpResponse
    status = response_started.get("status", 500)
    headers = response_started.get("headers", {})

    response = HttpResponse(
        bytes(response_body),
        status=status,
    )
    for key, value in headers.items():
        response[key] = value

    return response


class _ToolsetMethodCaller:

    def __init__(self, class_, method_name, context_kwarg, forward_context_kwarg):
        self.class_ = class_
        self.method_name = method_name
        self.context_kwarg = context_kwarg
        self.forward_context_kwarg = forward_context_kwarg

    def __call__(self, *args, **kwargs):
        # Get the class instance
        instance = self.class_(context=kwargs[self.context_kwarg],
                               request=django_request_ctx.get(SimpleNamespace()))
        # Get the method
        method = sync_to_async(_SyncToolCallWrapper(getattr(instance, self.method_name)))
        if not self.forward_context_kwarg:
            # Remove the context kwarg from kwargs
            del kwargs[self.context_kwarg]

        return method(*args, **kwargs)


MCP_SESSION_ID_HDR = "Mcp-Session-Id"


# FIXME: shall I reimplement the necessary without the
# Stuff pulled to support embedded server ?
class DjangoMCP(FastMCP):

    def __init__(self, name=None, instructions=None, stateless=False):
        # Prevent extra server settings as we do not use the embedded server
        super().__init__(name or "django_mcp_server", instructions)
        self.stateless = stateless
        engine = import_module(settings.SESSION_ENGINE)
        self.SessionStore = engine.SessionStore

    @property
    def session_manager(self) -> StreamableHTTPSessionManager:
        return StreamableHTTPSessionManager(
            app=self._mcp_server,
            event_store=self._event_store,
            json_response=True,
            stateless=True,  # Sessions will be managed as Django sessions.
        )

    def handle_django_request(self, request):
        """
        Handle a Django request and return a response.
        This method is called by the Django view when a request is received.
        """
        if not self.stateless:
            session_key = request.headers.get(MCP_SESSION_ID_HDR)
            if session_key:
                session = self.SessionStore(session_key)
                if session.exists(session_key):
                    request.session = session
                else:
                    return HttpResponse(status=404, content="Session not found")
            elif request.data.get('method') == 'initialize':
                # FIXME: Trick to read body before data to avoid DRF complaining
                request.session = self.SessionStore()
            else:
                return HttpResponse(status=400, content="Session required for stateful server")

        result = async_to_sync(_call_starlette_handler)(request, self.session_manager)

        # Only persist and strip the session in stateful mode when we actually
        # added it to the request.
        if not self.stateless and hasattr(request, "session"):
            request.session.save()
            result.headers[MCP_SESSION_ID_HDR] = request.session.session_key
            delattr(request, "session")

        return result

    def destroy_session(self, request):
        session_key = request.headers.get(MCP_SESSION_ID_HDR)
        if not self.stateless and session_key:
            self.SessionStore(session_key).flush()
            request.session = None

    def append_instructions(self, new_instructions):
        """
        Append instructions to the server instructions.
        This method is called by the Django view when a request is received.
        """
        inst = self._mcp_server.instructions
        if not inst:
            inst = new_instructions
        else:
            inst = inst.strip() + "\n\n" + new_instructions.strip()
        self._mcp_server.instructions = inst

    def register_mcptoolset(self, toolset):
        return toolset._add_tools_to(self._tool_manager)

    def register_drf_create_tool(
            self,
            view_class: type("GenericAPIView"),
            name=None,
            instructions=None,
            body_schema: dict | None = None,
            actions: dict | None = None,
    ):
        """
        Function or Decorator to register a DRF CreateModelMixin view as a MCP Toolset.
        :param view_class: The DRF view subclassing CreateModelMixin.
        :param name: the tool name, can be auto generated
        :param instructions: the instructions to provide to the MCP client, mandatory if the view does not have a docstring.
        :param body_schema: JSON Schema, optional in reasonably recent DRF that supports schema generation.
                            If DRF does not support schema generation, this becomes mandatory
        :param actions: DRF action mapping for ViewSet initialization. Omit if the class that is added is not a ViewSet
                        subclass. Example: {'post': 'create'}
        :return:
        """
        assert instructions or view_class.__doc__, "You need to provide instructions or the class must have a docstring"

        async def _dumb_create(body: dict):
            pass

        tool = self._tool_manager.add_tool(
            fn=_dumb_create,
            name=name or f"{view_class.__name__}_CreateTool",
            description=instructions or view_class.__doc__
        )
        tool.fn = sync_to_async(_DRFCreateAPIViewCallerTool(self, view_class, actions=actions))

        if body_schema is not None:
            tool.parameters['properties']['body'] = body_schema
        else:
            try:
                # Extract schema for a specific serializer manually
                tool.parameters['properties']['body'] = view_class.schema.map_serializer(view_class.serializer_class())
            except AttributeError:
                logger.warning(
                    "DRF version installed does not support schema generation, officially, trying privte API")
                try:
                    tool.parameters['properties']['body'] = view_class.schema._map_serializer(
                        view_class.serializer_class()
                    )
                except Exception:
                    logger.critical("DRF does not support schema generation, you must provide a body_schema parameter "
                                    "to the tool registration")
            except Exception:
                logger.critical(f"Error extracting schema for {view_class}, you must provide body_schema",
                                exc_info=True)
                raise

    def register_drf_list_tool(
            self,
            view_class: type("GenericAPIView"),
            name: str | None = None,
            instructions: str | None = None,
            actions: dict | None = None,
    ):
        assert instructions or view_class.__doc__, "You need to provide instructions or the class must have a docstring"

        async def _dumb_list():
            pass

        tool = self._tool_manager.add_tool(
            fn=_dumb_list,
            name=name or f"{view_class.__name__}_ListTool",
            description=instructions or view_class.__doc__
        )
        tool.fn = sync_to_async(_DRFListAPIViewCallerTool(self, view_class, actions=actions))

    def register_drf_update_tool(
            self,
            view_class: type("GenericAPIView"),
            name: str | None = None,
            instructions: str | None = None,
            body_schema: dict | None = None,
            actions: dict | None = None,
    ):
        """
        Function or Decorator to register a DRF CreateModelMixin view as a MCP Toolset.
        :param view_class: The DRF view subclassing CreateModelMixin.
        :param name: the tool name, can be auto generated
        :param instructions: the instructions to provide to the MCP client, mandatory if the view does not
                             have a docstring.
        :param body_schema: JSON Schema, optional in reasonably recent DRF that supports schema generation. If DRF does
                            not support schema generation, this becomes mandatory
        :param actions: DRF action mapping for ViewSet initialization. Omit if the class that is added is
                        not a ViewSet subclass. Example: {'put': 'update'}
        :return:
        """
        assert instructions or view_class.__doc__, "You need to provide instructions or the class must have a docstring"

        async def _dumb_update(id, body: dict):
            pass

        tool = self._tool_manager.add_tool(
            fn=_dumb_update,
            name=name or f"{view_class.__name__}_UpdateTool",
            description=instructions or view_class.__doc__
        )
        tool.fn = sync_to_async(_DRFUpdateAPIViewCallerTool(self, view_class, actions=actions))

        # Extract schema for a specific serializer manually
        if body_schema is not None:
            tool.parameters['properties']['body'] = body_schema
        else:
            try:
                # Extract schema for a specific serializer manually
                tool.parameters['properties']['body'] = view_class.schema.map_serializer(view_class.serializer_class())
            except AttributeError:
                logger.warning(
                    "DRF version installed does not support schema generation, officially, trying privte API")
                try:
                    tool.parameters['properties']['body'] = view_class.schema._map_serializer(
                        view_class.serializer_class()
                    )
                except Exception:
                    logger.critical("DRF does not support schema generation, you must provide a body_schema parameter "
                                    "to the tool registration")
            except Exception:
                logger.critical(f"Error extracting schema for {view_class}, you must provide body_schema",
                                exc_info=True)
                raise

    def register_drf_destroy_tool(
            self,
            view_class: type("GenericAPIView"),
            name: str | None = None,
            instructions: str | None = None,
            actions: dict | None = None,
    ):
        assert instructions or view_class.__doc__, "You need to provide instructions or the class must have a docstring"

        async def _dumb_delete(id):
            pass

        tool = self._tool_manager.add_tool(
            fn=_dumb_delete,
            name=name or f"{view_class.__name__}_DeleteTool",
            description=instructions or view_class.__doc__
        )
        tool.fn = sync_to_async(_DRFDeleteAPIViewCallerTool(self, view_class, actions=actions))


global_mcp_server = DjangoMCP(**getattr(settings, 'DJANGO_MCP_GLOBAL_SERVER_CONFIG', {}))


class ToolsetMeta(type):
    registry = {}

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        # Skip base class itself
        if name != "MCPToolset" and issubclass(cls, MCPToolset):
            ToolsetMeta.registry[name] = cls

    @staticmethod
    def iter_all():
        """
        Iterate over all toolsets
        """
        for name, cls in ToolsetMeta.registry.items():
            yield name, cls


class MCPToolset(metaclass=ToolsetMeta):
    """
    Base class for MCP toolsets. This class provides a way to create tools that can be used with
    the built in MCP serfver in a declarative way.

    ```
    class MyAppTools(MCPToolset):
        def my_tool(param : Type) -> ReturnType:
            ...
    ```

    Any "private" method (i.e. its name starting with _) will not be declared as a tool.
    Any other method is published as an MCP Tool that MCP Clients can use.

    During tool execution, self.request contains the original django request, this allows, for example,
    access to request.user ...

    """

    """You can define your own instance of DjangoMCP here """
    mcp_server: DjangoMCP = None

    def __init__(self, context=None, request=None):
        self.context = context
        self.request = request
        if self.mcp_server is None:
            self.mcp_server = global_mcp_server

    def _add_tools_to(self, tool_manager):
        """
        ADd tools to the manager
        :param tool_manager:
        :return: list of tools added
        """
        ret = []
        # ITerate all the methods whose name does not start with _ and register them with mcp_server.add_tool
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if not callable(method) or name.startswith("_"):
                continue
            tool = tool_manager.add_tool(sync_to_async(method))
            if tool.context_kwarg is None:
                forward_context = False
                tool.context_kwarg = "_context"
            else:
                forward_context = True
            tool.fn = _ToolsetMethodCaller(self.__class__, name, tool.context_kwarg, forward_context)
            ret.append(tool)
        return ret


def init():
    # Register the tools
    for _name, cls in ToolsetMeta.iter_all():
        if cls.mcp_server is None:
            cls.mcp_server = global_mcp_server

    for _name, cls in ToolsetMeta.iter_all():
        cls.mcp_server.register_mcptoolset(cls())

    from . import query_tool
    query_tool.init(global_mcp_server)


class _DRFRequestWrapper(HttpRequest):

    def __init__(self, mcp_server, mcp_request, method, body_json=None, id=None):
        super().__init__()
        self._serialized_body = json.dumps(body_json).encode("utf-8") if body_json else b''
        self.method = method
        self.content_type = "application/json"
        self.META = {
            'CONTENT_TYPE': 'application/json',
            'HTTP_ACCEPT': 'application/json',
            'CONTENT_LENGTH': len(self._serialized_body)
        }

        self._stream = BytesIO(self._serialized_body)
        self._read_started = False
        self.user = mcp_request.user
        self.session = mcp_request.session
        self.original_request = mcp_request
        self.path = f'/_djangomcpserver/{mcp_server.name}'
        if id:
            self.path += f"/{id}"


class BaseAPIViewCallerTool:
    view: Type["APIView"]

    @staticmethod
    def _patched_initialize_request(self, request, *args, **kwargs):
        original_request = request.original_request
        original_request.request = request
        original_request.method = request.method
        return original_request

    def __init__(self, view_class, **kwargs):
        view_class.initialize_request = self._patched_initialize_request
        self.view = view_class.as_view(**kwargs)


class _DRFCreateAPIViewCallerTool(BaseAPIViewCallerTool):
    def __init__(self, mcp_server, view_class, actions=None):
        if not issubclass(view_class, CreateModelMixin):
            raise ValueError(f"{view_class} must be a subclass of DRF CreateModelMixin")
        self.mcp_server = mcp_server
        self.view_class = view_class

        def raise_exception(exp):
            raise exp

        kwargs = dict(
            filter_backends=[],
            authentication_classes=[],
            permission_classes=view_class.permission_classes,
            handle_exception=raise_exception
        )
        if actions is not None:
            kwargs['actions'] = actions

        # Disable built in tauth
        super().__init__(view_class, **kwargs)

    def __call__(self, body: dict):
        # Create a request
        request = _DRFRequestWrapper(self.mcp_server, django_request_ctx.get(SimpleNamespace()), "POST", body_json=body)

        # Create the view
        try:
            return self.view(request).data
        except Exception as exp:
            logger.exception("Error in DRF tool invocation", exc_info=exp)
            raise exp


class _DRFListAPIViewCallerTool(BaseAPIViewCallerTool):
    def __init__(self, mcp_server, view_class, actions=None):
        if not issubclass(view_class, ListModelMixin):
            raise ValueError(f"{view_class} must be a subclass of DRF ListModelMixin")
        self.mcp_server = mcp_server
        self.view_class = view_class

        def raise_exception(exp):
            raise exp

        kwargs = dict(
            filter_backends=[],
            authentication_classes=[],
            permission_classes=view_class.permission_classes,
            handle_exception=raise_exception,
            pagination_class=None,
        )
        if actions is not None:
            kwargs['actions'] = actions

        # Disable built in tauth
        super().__init__(view_class, **kwargs)

    def __call__(self):
        # Create a request
        request = _DRFRequestWrapper(self.mcp_server, django_request_ctx.get(SimpleNamespace()), "GET")

        # Create the view
        try:
            return self.view(request).data
        except Exception as exp:
            logger.exception("Error in DRF tool invocation", exc_info=exp)
            raise exp


class _DRFUpdateAPIViewCallerTool(BaseAPIViewCallerTool):
    def __init__(self, mcp_server, view_class, actions=None):
        if not issubclass(view_class, UpdateModelMixin):
            raise ValueError(f"{view_class} must be a subclass of DRF UpdateModelMixin")
        self.mcp_server = mcp_server
        self.view_class = view_class

        def raise_exception(exp):
            raise exp

        kwargs = dict(
            filter_backends=[],
            authentication_classes=[],
            permission_classes=view_class.permission_classes,
            handle_exception=raise_exception
        )
        if actions is not None:
            kwargs['actions'] = actions

        # Disable built in tauth
        super().__init__(view_class, **kwargs)

    def __call__(self, id, body: dict):
        # Create a request
        request = _DRFRequestWrapper(self.mcp_server, django_request_ctx.get(SimpleNamespace()), "PUT", id=id,
                                     body_json=body)

        # Create the view
        try:
            return self.view(request, **{(self.view_class.lookup_url_kwarg or self.view_class.lookup_field): id}).data
        except Exception as exp:
            logger.exception("Error in DRF tool invocation", exc_info=exp)
            raise exp


class _DRFDeleteAPIViewCallerTool(BaseAPIViewCallerTool):
    def __init__(self, mcp_server, view_class, actions=None):
        if not issubclass(view_class, DestroyModelMixin):
            raise ValueError(f"{view_class} must be a subclass of DRF DestroyModelMixin")
        self.mcp_server = mcp_server
        self.view_class = view_class

        def raise_exception(exp):
            raise exp

        kwargs = dict(
            filter_backends=[],
            authentication_classes=[],
            permission_classes=view_class.permission_classes,
            handle_exception=raise_exception
        )
        if actions is not None:
            kwargs['actions'] = actions

        # Disable built in tauth
        super().__init__(view_class, **kwargs)

    def __call__(self, id):
        # Create a request
        request = _DRFRequestWrapper(self.mcp_server, django_request_ctx.get(SimpleNamespace()), "DELETE", id=id)

        # Create the view
        try:
            return self.view(request, **{(self.view_class.lookup_url_kwarg or self.view_class.lookup_field): id}).data
        except Exception as exp:
            logger.exception("Error in DRF tool invocation", exc_info=exp)
            raise exp


def drf_publish_create_mcp_tool(
        *args,
        name: str | None = None,
        instructions: str | None = None,
        server: DjangoMCP | None = None,
        body_schema: dict | None = None,
        actions: dict | None = None,
):
    """
    Function or Decorator to register a DRF CreateModelMixin view as an MCP Toolset.

    :param instructions: Instructions to provide to the MCP client.
    :param server: The server to use, if not set, the global one will be used.
    :param body_schema: JSON Schema, optional in reasonably recent DRF that supports schema generation.
                        If DRF does not support schema generation, this becomes mandatory
    :param actions: DRF action mapping for ViewSet initialization. Omit if the class that is added is not a
                    ViewSet subclass. Example: {'post': 'create'}

    :return:
    """
    assert len(args) <= 1, "You must provide the DRF view or nothing as argument"

    def decorator(view_class):
        (server or global_mcp_server).register_drf_create_tool(
            view_class,
            name=name,
            instructions=instructions,
            body_schema=body_schema,
            actions=actions,
        )
        return view_class

    if args:
        decorator(args[0])
    else:
        return decorator


def drf_publish_list_mcp_tool(
        *args,
        name: str | None = None,
        instructions: str | None = None,
        server: DjangoMCP | None = None,
        actions: dict | None = None):
    """
    Function or Decorator to register a DRF ListModelMixin view as an MCP Toolset.

    :param instructions: Instructions to provide to the MCP client.
    :param server: The server to use, if not set, the global one will be used.
    :param actions: DRF action mapping for ViewSet initialization. Omit if the class that is added is not a
                    ViewSet subclass. Example: {'get': 'list'}
    :return:
    """
    assert len(args) <= 1, "You must provide the DRF view or nothing as argument"

    def decorator(view_class):
        (server or global_mcp_server).register_drf_list_tool(
            view_class,
            name=name,
            instructions=instructions,
            actions=actions,
        )
        return view_class

    if args:
        decorator(args[0])
    else:
        return decorator


def drf_publish_update_mcp_tool(
        *args,
        name: str | None = None,
        instructions: str | None = None,
        server: DjangoMCP | None = None,
        body_schema: dict | None = None,
        actions: dict | None = None,
):
    """
    Function or Decorator to register a DRF UpdateModelMixin view as an MCP Toolset.

    :param instructions: Instructions to provide to the MCP client.
    :param server: The server to use, if not set, the global one will be used.
    :param body_schema: JSON Schema, optional in reasonably recent DRF that supports schema generation. If DRF does not
                        support schema generation, this becomes mandatory
    :param actions: DRF action mapping for ViewSet initialization. Omit if the class that is added is not a ViewSet
                    subclass. Example: {'put': 'update'}'}
    :return:
    """
    assert len(args) <= 1, "You must provide the DRF view or nothing as argument"

    def decorator(view_class):
        (server or global_mcp_server).register_drf_update_tool(
            view_class,
            name=name,
            instructions=instructions,
            body_schema=body_schema,
            actions=actions,
        )
        return view_class

    if args:
        decorator(args[0])
    else:
        return decorator


def drf_publish_destroy_mcp_tool(
        *args,
        name: str | None = None,
        instructions: str | None = None,
        server: DjangoMCP | None = None,
        actions: dict | None = None,
):
    """
    Function or Decorator to register a DRF UpdateModelMixin view as an MCP Toolset.

    :param instructions: Instructions to provide to the MCP client.
    :param server: The server to use, if not set, the global one will be used.
    :param actions: DRF action mapping for ViewSet initialization. Omit if the class that is added is not a ViewSet
                    subclass. Example: {'delete': 'destroy'}'}
    :return:
    """
    assert len(args) <= 1, "You must provide the DRF view or nothing as argument"

    def decorator(view_class):
        (server or global_mcp_server).register_drf_destroy_tool(
            view_class,
            name=name,
            instructions=instructions,
            actions=actions,
        )
        return view_class

    if args:
        decorator(args[0])
    else:
        return decorator



from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules


class McpServerConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "mcp_server"

    def ready(self):
        autodiscover_modules('mcp')
        from .djangomcp import init
        init()



import asyncio
import os
from io import TextIOWrapper

from django.core.management.base import BaseCommand
from asgiref.sync import async_to_sync
import anyio
from mcp import ClientSession, StdioServerParameters, stdio_server
from mcp.client.stdio import stdio_client

from mcp_server import mcp_server


async def _check_client():

    cl_wr, svr_rd = anyio.create_memory_object_stream(0)
    svr_wr, cl_rd = anyio.create_memory_object_stream(0)

    async def run_server():
        await mcp_server._mcp_server.run(
            svr_rd,
            svr_wr,
            mcp_server._mcp_server.create_initialization_options(),
        )

    async def run_client():
        async with ClientSession(cl_rd, cl_wr) as session:
            await session.initialize()

            print("Tools discovered in server:")
            tool_list = await session.list_tools()
            for tool in tool_list.tools:
                print(f'\n\t{tool.name}: {tool.description}')
                print(f'\tParameters: {tool.inputSchema}')

            print("Resources discovered in server:")
            resource_list = await session.list_resources()
            for resource in resource_list.resources:
                print(f'\n\t{resource.name}: {resource.description}')

            print("Prompts discovered in server:")
            prompt_list = await session.list_prompts()
            for prompt in prompt_list.prompts:
                print(f'\n\t{prompt.name}: {prompt.description}')

    async with anyio.create_task_group() as tg:
        tg.start_soon(run_server)
        tg.start_soon(run_client)


_check_client_sync = async_to_sync(_check_client)


class Command(BaseCommand):
    help = 'Inspect installed tools, resources and prompts'

    def handle(self, *args, **options):
        _check_client_sync()
